/* mnemon - A memory training library
 *
 * Copyright © 2006,2011 Carl Worth
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA."
 */

#ifndef MNEMON_H_INCLUDED
#define MNEMON_H_INCLUDED

#define unused(foo) foo __attribute__((unused))

typedef int bool_t;

typedef struct _item {
    int score;
    char *challenge;
    char *response;
} item_t;

typedef struct _bin {
    int score;
    int items_size;
    int num_items;
    item_t **items;
} bin_t;

typedef enum {
    CATEGORY_ORDER_RANDOM,
    CATEGORY_ORDER_SEQUENTIAL
} category_order_t;

typedef enum {
    CHALLENGE_TYPE_TEXT,
    CHALLENGE_TYPE_IMAGE,
    CHALLENGE_TYPE_AUDIO,
    CHALLENGE_TYPE_MIDI,
    CHALLENGE_TYPE_TEXT_TO_SPEECH
} challenge_type_t;

typedef struct _category {
    char *name;
    int items_size;
    int num_items;
    item_t *items;

    /* Support sequential introduction of items from bin 0 */
    category_order_t order;
    /* Support categories where responses are timed (0.0 == disable). */
    double time_limit;
    int bin_zero_head;
    /* Support challenges of non-text types (image, audio, etc.) */
    challenge_type_t challenge_type;
    /* Whether to repeat afterwards (for a little extra reinforcement) */
    bool_t repeat;
} category_t;

typedef struct _mnemon {
    char *dir_name;

    int categories_size;
    int num_categories;
    category_t *categories;

    int bins_size;
    int num_bins;
    bin_t *bins;
} mnemon_t;

/* Initialize a new mnemon object. This function must be called before
 * any other mnemon functions are used. */
void
mnemon_init (mnemon_t *mnemon);

/* Inidicate the caller is finished with a mnemon object. Free all
 * resources associated with this object. After this call, the given
 * mnemon object should not be passed to any other menmon function,
 * (except mnemon_init to start over). */
void
mnemon_fini (mnemon_t *mnemon);

/* Load a specific category of mnemon challenges. The name should
 * indicate the name of a file within the user's .mnemon directory. */
void
mnemon_load_category (mnemon_t		*mnemon,
		      const char	*name);

/* Load all categories of mnemon challenges.
 *
 * This is equivalent to calling mnemon_load_category for all files
 * found within the user's .mnemon directory. */
void
mnemon_load (mnemon_t *mnemon);

/* Select the next (weighted) random item to challenge the user.
 *
 * This function provides four return values (and yes, that's
 * exceedingly awkward and a simpler interface should be designed to
 * replace this):
 *
 *	bin:		The bin from which the item was selected
 *
 *	item_index:	The index within the bin of the slected item
 *
 *	category:	The name of the category for this item
 *
 *	introduced:     A flag indicating whether this is a newly
 *			introduced item. Items from bin 0 always count
 *			as newly introduced. If there is no bin 0,
 *			then items from the lowest non-negative bin
 *			will be flagged as introduced.
 *
 * The selection system is designed to rapidly reinforce items needing
 * to be learned and provide exponentially less reinforcement for
 * items as mastery is displayed. This is achieved by storing the
 * items in a series of numbered bins.
 *
 * Items start in bin 0 indicating that they have never been presented
 * to a user. When an item is presented to the user and answered
 * correctly, it is moved into the bin of the next higher number.
 *
 * However, when an item is answered incorrectly, it is moved directly
 * to bin -2 (if coming from a bin of a positive number), or the bin
 * of the next lower integer (more negative) if coming from a bin of a
 * negative number.
 *
 * When selecting a new item to challenge, first a bin is chosen
 * (considering only the non-empty bins). The bin with the lowest
 * number is the most likely to be chosen, while each succesively-
 * higher-numbered bin has a probability one-half of that of the
 * previous bin.
 */
void
mnemon_select_item (mnemon_t	 *mnemon,
		    bin_t	**bin_ret,
		    int		 *item_index_ret,
		    category_t	**category_ret,
		    int	         *introduced_ret);

/* Update an item based on a user's answer (correct or incorrect).
 *
 * The bin and item_index should be exactly as returned by
 * mnemon_select_item.  The correct flag should indicate whether the
 * user answered the challenge correctly or not, (and should only
 * count as correct if the answer was within the time limit, if any).
 *
 * The item will be moved from its current bin to a new bin based on
 * whether the challenge was answered correctly. The bin updates are
 * as follows:
 *
 * If the answer was correct:
 *	Increase the bin number by 1
 *	If the new bin number is 0, set it to 1 (0 bin is for new items)
 *
 * If the answer was incorrect:
 *	If the old bin was positive, move to bin -2 (for extra training)
 *	Otherwise decrease the bin number by 1
 *
 * Note: All item and bin movement is kept only in memory. In order to
 * save this progress to disk, the caller must call mnemon_save.
 */
void
mnemon_score_item (mnemon_t *mnemon,
		   bin_t *bin,
		   unsigned int item_index,
		   bool_t correct);

/* Save the user's progress by updating the category files in the
 * users .mnemon directory. */
void
mnemon_save (mnemon_t *mnemon);

/* Remove a bin of a particular number.
 *
 * This can be useful in situations such as wanting to practice
 * mastery of learned items without mixing in new items that the user
 * has never seen before. This could be achieved by removing bin 0,
 * for example.
 */
void
mnemon_remove_bin (mnemon_t *mnemon, int bin_number);

/* Find the category to which an item belongs. */
category_t *
mnemon_item_category (mnemon_t	*mnemon,
		      item_t	*item);

/* Get a category by name if it exists */
category_t *
mnemon_get_category_if_exists (mnemon_t	    *mnemon,
			       const char   *name);


#endif
